/******************************************************************************
*
* project name:    TICHESS
* project version: 2.00
* file name:       tichess.c
* initial date:    20/02/2000
* author:          thomas.nussbaumer@peem.com
* description:     the entry point and the splashscreen stuff
*
* -----------------------------------------------------------------------------
* Revision History
* -----------------------------------------------------------------------------
*
* v1.15: 08/04/2000
*        comment about failure in DOORS II 0.95 added
*        
* v1.14: 01/04/2000
*        changes due to new version TIGCC standard library 1.5
*        
* v1.13: 24/03/2000
*        fatal bug which crashes calculator fixed
*        
* v1.12: 24/03/2000
*        special clear screen at startup added
*        
* v1.11: 22/03/2000
*        using compiler define DEVELOPER_VERSION to toggle special functions
*        
* v1.10: 20/03/2000
*        grayscale/monochrome now dynamically switchable
*        handling of saved configuration added
*        
* v1.09: 11/03/2000
*        level 0 added to give beginners a change ;-)
*        
* v1.08: 10/03/2000
*        monochrome version support added
*
* v1.07: 10/03/2000
*        Splashscreen now monochrome
*        using NEW gray.o and gray.h (HW autodetect and no problems on HW2)
*
* v1.06: 08/03/2000
*        HW checking routine from Julien Muchembled integrated
*        now showing web address on splashscreen
*
* v1.05: 07/03/2000
*        (1) modifications for DOORS version
*        (2) modifications for batch building
*        (3) splashscreen modified
*        (4) some modifications for the dialogs
*
******************************************************************************/

//-----------------------------------------------------------------------------
// this define is for batch compiling where the defines are specified
// at the commandline
//-----------------------------------------------------------------------------
#ifndef AUTO_BUILD

//=============================================================================
// set the following define to 1 if you want the DOORS OS version
//
// NOTE: Due to the fact that no libraries from DOORS are used there is not 
//       much difference even in size between the DOORS and the NOSTUB version.
//       The complete development and test cycles are done with the NOSTUB
//       version. So there is no garantee that the DOORS version will work
//       completely correct.
//
// NOTE2: If TI-Chess is build for DOORS, you have to use DOORS II 0.96 or
//        greater, because DOORS II 0.95 contains a terrible bug within
//        variable relocation which will crash your calculator !!!
//=============================================================================
#define COMPILE_FOR_DOORS 0

//=============================================================================
// THE FOLLOWING PRECOMPILER DEFINES ARE USED TO SET A SPECIFIC CALCULATOR TYPE
//
// !!!!!! NOTE: ONLY ONE OF THE DEFINES MAY BE SET TO '1' !!!!!
//=============================================================================
#define TI89_VERSION      1    // generate version for TI89
#define TI92_VERSION      0    // generate version for TI92

#endif

#if COMPILE_FOR_DOORS
   #include <doors.h>
#else
   #include <nostub.h>
#endif
#include <graph.h>
#include <kbd.h>
#include <alloc.h>
#include <vat.h>
#include <string.h>
#include <args.h>
#include <ctype.h>
#include <setjmp.h>
#include <gray.h>

#include "version.h"
#include "tichess.h"
#include "logout.h"      // contains macros for log writing to link port


//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// due to a very strange behavior of TIGCC we cannot use separate object until
// yet (the globals are not initialized correctly - NOT even the constants !!)
//
// So we've to go that really silly "include all into one" way ...
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#include "loadsave.c" // loading/saving stuff
#include "gui.c"      // gui stuff
#include "engine.c"   // engine stuff
#include "hash.c"     // hashtable stuff
#include "menu.c"     // menu stuff

#if ENABLE_LOGGING
#include "logout.c"   // includes logging routines if necessary
#endif


#if TI92_VERSION
int _ti92plus;        // needed to produce ti92plus code
#else 
int _ti89;            // needed to produce ti89 code
#endif

//--------------------------------------------------
// declarations of the necessary "external" routines
//--------------------------------------------------
extern char default_hashmode;
extern int  nr_planes;

int  InitMemory();
void FreeMemory();
void MainGUI();
int  LoadDefaults();


/*===========================================================================*/
/* special effect clear screen                                               */
/*===========================================================================*/
void FXClearScreen() {
    int height  = ScrRect->xy.y1;
    int height2 = height/2;
    int width   = ScrRect->xy.x1;
    int width2  = width/2;
    int i,j;

    //---------------------
    // clear in y direction
    //---------------------
    for (i=0;i<=height2;i++) {
        DrawLine(0,i,width,i,A_REPLACE);
        DrawLine(0,height-i,width,height-i,A_REPLACE);
        DrawLine(0,i-1,width,i-1,A_REVERSE);
        DrawLine(0,height-i+1,width,height-i+1,A_REVERSE);
        for (j=0;j<=10000;j++);
    }
    
    //-----------------------------
    // clear in x direction
    //-----------------------------
    for (i=0;i<=width2;i++) {
        DrawPix(i,height2,A_REVERSE);
        DrawPix(width-i,height2,A_REVERSE);
        for (j=0;j<=3000;j++);
    }
}


/*===========================================================================*/
/* draws the splashscreen                                                    */
/*                                                                           */
/* NOTE: not grayscale                                                       */
/*===========================================================================*/
void SplashScreen(void) {
    int maxx    = ScrRect->xy.x1;
    int maxy    = ScrRect->xy.y1;
#if BETA_VERSION
    char*    s1 = "BETA RELEASE";
    char*    s2 = "don't distribute";
#else
    char*    s1 = "The TI-Chess Team";
    char*    s2 = "proudly presents";
#endif
    char*    s3 = "TI-Chess v"TIC_VERSION_MAIN""TIC_VERSION_SUB;
    char*    s4 = "http://members.xoom.com/tnussb2";

    int      px1,px2,px3,px4;
    int      py1,py2,py3,py4;

    px1 = (maxx-DrawStrWidth(s1,F_8x10))/2;
    px2 = (maxx-DrawStrWidth(s2,F_6x8))/2;
    px3 = (maxx-DrawStrWidth(s3,F_8x10))/2;
    px4 = (maxx-DrawStrWidth(s4,F_4x6))/2;
    py1 = maxy/4;
    py2 = maxy/2;
    py3 = maxy*3/4;
    py4 = maxy-5;

    FontSetSys(F_8x10);
    DrawStr(px1,py1,s1,A_REPLACE);
    DrawStr(px3,py3,s3,A_REPLACE);
    FontSetSys(F_6x8);
    DrawStr(px2,py2,s2,A_REPLACE);
    FontSetSys(F_4x6);
    DrawStr(px4,py4,s4,A_REPLACE);

    ngetchx();
}


/*===========================================================================*/
/* MAIN - where all the fun starts ...                                       */
/*===========================================================================*/
void _main(void) {
    int          i;
    char*        s1 = "ERROR: Memory";
    char*        s2 = "Insufficient Memory.";
    char         s3[30];
    unsigned int size;
#ifndef DOORS
    LCD_BUFFER buffer;
    LCD_save(buffer);
#endif

    LoadDefaults();   // load defaults from ticcfg

    default_hashmode = defaults.hashtables;

    if (defaults.graphics == OPTION_ON) nr_planes = 2;
    else                                nr_planes = 1;
    
    FXClearScreen();
    
#if BETA_VERSION
    if (default_hashmode == OPTION_ON) size = sizeof(hentry_t)*HASHSIZE*2;
    else                               size = 0;

    size += sizeof(char)*BOARD_SIZE + \
       sizeof(int)*MAX_DEPTH+sizeof(move_t)*(STACK_SIZE+1) + \
       sizeof(int)*MAX_DEPTH+sizeof(killer_t)*MAX_DEPTH + \
       sizeof(bicolor_t)*BOARD_SIZE + \
       sizeof(bicolor_t)*(H_COLUMN+2)*2 + \
       sizeof(int)*MAX_DEPTH*4 + \
       sizeof(fromto_t*)*MAX_DEPTH + \
       sizeof(fromto_t)*MAX_DEPTH*MAX_DEPTH + \
       sizeof(hash_t*)*12 + \
       sizeof(hash_t)*64*12 + \
       sizeof(store_t)*MAX_MOVES_STORED;
#endif
    FontSetSys(F_6x8);

    if (!InitMemory()) {
        FreeMemory();

        size = sizeof(char)*BOARD_SIZE + \
           sizeof(int)*MAX_DEPTH+sizeof(move_t)*(STACK_SIZE+1) + \
           sizeof(int)*MAX_DEPTH+sizeof(killer_t)*MAX_DEPTH + \
           sizeof(bicolor_t)*BOARD_SIZE + \
           sizeof(bicolor_t)*(H_COLUMN+2)*2 + \
           sizeof(int)*MAX_DEPTH*4 + \
           sizeof(fromto_t*)*MAX_DEPTH + \
           sizeof(fromto_t)*MAX_DEPTH*MAX_DEPTH + \
           sizeof(hash_t*)*12 + \
           sizeof(hash_t)*64*12 + \
           sizeof(store_t)*MAX_MOVES_STORED;

        sprintf(s3,"(Req'd %u Bytes)",size); 

        FontSetSys(F_8x10);
        DrawStr((ScrRect->xy.x1-DrawStrWidth(s1,F_8x10))/2,ScrRect->xy.y1/3,s1,A_REPLACE);
        FontSetSys(F_6x8);
        DrawStr((ScrRect->xy.x1-DrawStrWidth(s2,F_6x8))/2,ScrRect->xy.y1/2,s2,A_REPLACE);
        DrawStr((ScrRect->xy.x1-DrawStrWidth(s3,F_6x8))/2,ScrRect->xy.y1*2/3,s3,A_REPLACE);
        ngetchx();
#ifndef DOORS
        LCD_restore(buffer);
#endif       
        return;
    }

#if BETA_VERSION
    sprintf(s3,"[using %u Bytes]",size); 
    FontSetSys(F_6x8);
    DrawStr(0,0,s3,A_REPLACE);
#endif

    SplashScreen();
    FXClearScreen();

    if (nr_planes == 2) {
        if (!GrayMode(GRAY_ON)) {
            nr_planes = 1;
            SaveDefaults();
#ifndef DOORS
            LCD_restore(buffer);
#endif
            return;
        }
    }

    MainGUI();

    if (nr_planes == 2) GrayMode(GRAY_OFF);
     
    FreeMemory();
    FXClearScreen();

#ifndef DOORS
    LCD_restore(buffer);
#endif
}

//##########################################################################
//###################### NO MORE FAKES BEYOND THIS LINE ####################
//##########################################################################
